/*
 * Copyright (c) 2014 Lutz Kaemmerer, 2015 Toni Volkmer
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <complex.h>
#include <math.h>

#ifdef __linux__
  /* for getcurtime() */
  #include <sys/time.h>
#else
  /* for clock() */
  #include <time.h>
#endif

/* check for ctrl-c interrupt via utIsInterruptPending() */
#if defined (_WIN32)
  #include <windows.h>
#elif defined (__linux__)
  #include <unistd.h>
#endif

#include "mex.h"

#if __STDC_VERSION__ >= 199901L
/* C99 code */
#define C99
  #define STATIC static
  #define INLINE inline
#else
  #define STATIC
  #define INLINE
#endif

#ifdef _OPENMP
#include <omp.h>
#endif

/* Display constants, default is verbose */
#define DISPLAY_OFF 0
#define DISPLAY_XS 5
#define DISPLAY_ITER 10
#define DISPLAY_VERBOSE 255
#define DISPLAY_INTERVAL 30.0

#define OMP_M_INCR 100000

#ifdef HAVE_OCTAVE
int utIsInterruptPending() { return 0; };
#include <octave/quit.h>
#else
extern int utIsInterruptPending();
#define OCTAVE_QUIT
#endif

/* if enabled, use 64-bit map instead of 32-bit map */

/* if enabled, check for correctness and return error to matlab if check fails */
#undef  LSIZEREDUCE_USE_ASSERT


  #define GET_BIT(bitmap, index) get_bit_64(bitmap, index)
  #define SET_BIT(bitmap, index) set_bit_64(bitmap, index)

#ifdef LSIZEREDUCE_USE_ASSERT
  #warning using assert
  static void assert_lsizereduce_fail(char *cond, const char *file, unsigned int line)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex:assert", "assert failed in %s at line %d: %s", file, line, cond);
  }
  #define assert_lsizereduce(condition) {if (!(condition)) assert_lsizereduce_fail(__STRING(condition), __FILE__, __LINE__);}
#endif

/**
  * Returns current time in seconds for time measurements.
  * On Linux, the current system time (gettimeofday()) is used,
  * otherwise the processor time used by the current program (clock()).
  */
STATIC INLINE double getcurtime(void)
{
#ifdef __linux__
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return 1.0 * tv.tv_sec + 1e-6 * tv.tv_usec;
#else
  return (double) clock() / CLOCKS_PER_SEC;
#endif
}

/**
  * Calculates rem = a mod b, where 0 <= rem < b.
  */
STATIC INLINE uint64_t modl(int64_t a, int64_t b)
{
  int64_t rem = a % b;
  if (rem < 0)
    rem += b;
#ifdef GENLATTICE_USE_ASSERT
  assert_genlattice(rem >= 0 && rem < b);
#endif

  return (uint64_t) rem;
}

/**
  * Gets bit at certain index from bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE uint64_t get_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1ULL << toffset;

  return temp[tindex] & tbit;
}

/**
  * Sets bit at certain index in bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE void set_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1ULL << toffset;

  temp[tindex] |= tbit;
}

STATIC INLINE int64_t min_i(int64_t a, int64_t b)
{
  return a < b ? a : b;
}

STATIC INLINE int64_t max_i(int64_t a, int64_t b)
{
  return a >= b ? a : b;
}

STATIC INLINE int32_t min_i32(int32_t a, int32_t b)
{
  return a < b ? a : b;
}

static void flushMatlab(void)
{
  fflush(NULL);
/*  mexEvalString("drawnow;"); */
}

/* Adapted from http://www.algorithmist.com/index.php/Prime_Sieve_of_Eratosthenes.c */
STATIC INLINE uint32_t prime_sieve_get(uint32_t *sieve, uint64_t b)
{
  return (sieve[(b)>>5]>>((b)&31))&1;
}

static void prime_sieve(uint32_t *sieve, uint32_t P1, uint32_t P2, uint32_t P3)
{
  uint32_t i, j, k;
  memset(sieve, 0, P1*sizeof(uint32_t));
  for (k = 1; k <= P3; k++)
    if (prime_sieve_get(sieve, k)==0)
      for(j=2*k+1,i=2*k*(k+1);i<P2;i+=j)
        sieve[i>>5]|=1<<(i&31);
}
STATIC INLINE int isprime(uint32_t *sieve, int64_t p)
{
  return p==2 || (p>2 && (p&1)==1 && (prime_sieve_get(sieve, (p-1)>>1)==0));
}

/**
  */
static int64_t lattice_size_reduce(int d, int64_t *z, int *freq, int len_freq, int64_t M_input, int64_t Mmin, int prime, int nthreads, int display)
{
  int t, k;

  const uint64_t bitmap_length = 1+M_input/(1U<<6);
  const uint64_t bitmap_size = bitmap_length * sizeof(uint64_t);
  int64_t *tempsp = (int64_t *) mxMalloc(len_freq * sizeof(int64_t));
  uint32_t *sieve;
  int64_t curM, Mstart, lastSuccess = 0;
  double time0, time1, time2;

  volatile bool interrupted = false;
  volatile bool found = false;
 
  time0 = getcurtime();

  if (prime == 1)
  {
    if (display >= DISPLAY_VERBOSE)
    {
      printf("Searching for prime numbers ...\n");
      flushMatlab();
    }

    sieve = (uint32_t *) mxMalloc((M_input/64 + 1) * sizeof(uint32_t));
    prime_sieve(sieve, M_input/64 + 1, lrint(ceil(M_input/2.0)), lrint(ceil(ceil(sqrt((double) M_input))/2.0)));
  }

  if (display >= DISPLAY_VERBOSE)
  {
    printf("Calculating inner product freq * z ...\n");
    flushMatlab();
  }

  for (k = 0; k < len_freq; k++)
  {
    tempsp[k] = ((int64_t) freq[d*k]) * z[0];
    for (t = 1; t < d; t++)
      tempsp[k] += ((int64_t) freq[d*k+t]) * z[t];
  }

  if (display >= DISPLAY_VERBOSE)
  {
    printf("Starting search for smaller lattice size M ...\n");
    flushMatlab();
  }

  time1 = time0;

for (Mstart = M_input; Mstart >= Mmin && lastSuccess == 0; Mstart -= OMP_M_INCR)
{
  int64_t innerLoopLowerM = max_i(max_i(Mmin, Mstart - OMP_M_INCR + 1), 1);
 #pragma omp parallel shared(found,interrupted) private(t,k) num_threads(nthreads)
 {
  uint64_t *bitmap;
  #pragma omp critical
  {
    bitmap = (uint64_t *) mxMalloc(bitmap_size);
  }

/*for (curM = M_input; curM >= Mmin; curM--)
    for (curM = Mstart; curM >= Mmin && curM > Mstart - OMP_M_INCR; curM--)
 */
  #pragma omp for schedule(dynamic,1)
  for (curM = Mstart; curM >= innerLoopLowerM; curM--)
  {
    uint64_t cur_bitmap_size;

    if (interrupted)
      continue;

#ifdef _OPENMP
   if (omp_get_thread_num() == 0)
#endif
   {
    OCTAVE_QUIT;
    if (utIsInterruptPending() != 0) {
      #pragma omp critical
      {
        mexWarnMsgTxt("CTRL-C detected. Aborting...");
        flushMatlab();

        interrupted = true;
      }
      continue;
    }
   }

    if (prime == 1 && isprime(sieve, curM) == 0)
      continue;

#ifdef _OPENMP
   if (omp_get_thread_num() == 0)
#endif
   {
    if (display >= DISPLAY_VERBOSE || display == DISPLAY_XS)
    {
      #pragma omp critical
      {
        time2 = getcurtime();
        if (time2 - time1 > DISPLAY_INTERVAL)
        {
          time1 = time2;
          if (lastSuccess == 0)
  	    printf("curr candidate: %ld\n", curM);
          else
            printf("last success: %ld, curr candidate: %ld\n", lastSuccess, curM);
          flushMatlab();
        }
      }
    }
   }

    cur_bitmap_size = (1+curM/(1U<<6)) * sizeof(uint64_t);
    memset(bitmap, 0, cur_bitmap_size);

    for (k = 0; k < len_freq; k++)
    {
      uint64_t index;
      index = modl(tempsp[k], curM);
        
      if (GET_BIT(bitmap, index) != 0)
        break;
      SET_BIT(bitmap, index);
    }

    if (k == len_freq)
    {
      #pragma omp critical
      {
        if (found)
	{
	  if (curM < lastSuccess)
            lastSuccess = curM;
	}
	else
          lastSuccess = curM;

	found = true;

        if (display >= DISPLAY_ITER)
        {
#ifdef _OPENMP
          printf("thread %d found lattice size: %ld\n", omp_get_thread_num(), curM);
#else
          printf("found lattice size: %ld\n", curM);
#endif
          flushMatlab();
          time1 = getcurtime();
        }
      }
    }
  }
  #pragma omp critical
  {
    mxFree(bitmap);
  }
 }

  if (interrupted)
  {
    mxFree(tempsp);
    if (prime == 1)
      mxFree(sieve);

    return -1;
  }
}

  mxFree(tempsp);
  if (prime == 1)
    mxFree(sieve);

  if (interrupted)
    return -1;

  if (display >= DISPLAY_ITER)
  {
    printf("Time: %.3e\n", getcurtime() - time0);
  }

  return lastSuccess;
}

/**
  */
static int64_t lattice_size_increase(int d, int64_t *z, int *freq, int len_freq, int64_t M_input, int64_t Mmin, int prime, int nthreads, int display)
{
  int t, k;

  const uint64_t bitmap_length = 1+M_input/(1U<<6);
  const uint64_t bitmap_size = bitmap_length * sizeof(uint64_t);
  int64_t *tempsp = (int64_t *) mxMalloc(len_freq * sizeof(int64_t));
  uint32_t *sieve;
  int64_t curM, Mstart, lastSuccess = 0;
  double time0, time1, time2;

  volatile bool interrupted = false;
  volatile bool found = false;
 
  time0 = getcurtime();

  if (prime == 1)
  {
    if (display >= DISPLAY_VERBOSE)
    {
      printf("Searching for prime numbers ...\n");
      flushMatlab();
    }

    sieve = (uint32_t *) mxMalloc((M_input/64 + 1) * sizeof(uint32_t));
    prime_sieve(sieve, M_input/64 + 1, lrint(ceil(M_input/2.0)), lrint(ceil(ceil(sqrt((double) M_input))/2.0)));
  }

  if (display >= DISPLAY_VERBOSE)
  {
    printf("Calculating inner product freq * z ...\n");
    flushMatlab();
  }

  for (k = 0; k < len_freq; k++)
  {
    tempsp[k] = ((int64_t) freq[d*k]) * z[0];
    for (t = 1; t < d; t++)
      tempsp[k] += ((int64_t) freq[d*k+t]) * z[t];
  }

  if (display >= DISPLAY_VERBOSE)
  {
    printf("Starting search for smaller lattice size M ...\n");
    flushMatlab();
  }

  time1 = time0;

for (Mstart = Mmin; Mstart <= M_input && lastSuccess == 0; Mstart += OMP_M_INCR)
{
  int64_t innerLoopUpperM = min_i(M_input, Mstart + OMP_M_INCR - 1);
 #pragma omp parallel shared(found,interrupted) private(t,k) num_threads(nthreads)
 {
  uint64_t *bitmap;
  #pragma omp critical
  {
    bitmap = (uint64_t *) mxMalloc(bitmap_size);
  }

/*for (curM = Mmin; curM <= M_input; curM++)
    for (curM = Mstart; curM <= M_input && curM < Mstart + OMP_M_INCR; curM++)
 */
  #pragma omp for schedule(dynamic,1)
  for (curM = Mstart; curM <= innerLoopUpperM; curM++)
  {
    uint64_t cur_bitmap_size;

    if (found || interrupted)
      continue;

#ifdef _OPENMP
    if (omp_get_thread_num() == 0)
#endif
    {
      OCTAVE_QUIT;
      if (utIsInterruptPending() != 0) {
        #pragma omp critical
	{
          mexWarnMsgTxt("CTRL-C detected. Aborting...");
          flushMatlab();

          interrupted = true;
        }
        continue;
      }
    }

    if (prime == 1 && isprime(sieve, curM) == 0)
      continue;

#ifdef _OPENMP
    if (omp_get_thread_num() == 0)
#endif
    {
      if (display >= DISPLAY_VERBOSE || display == DISPLAY_XS)
      {
        #pragma omp critical
	{
          time2 = getcurtime();
          if (time2 - time1 > DISPLAY_INTERVAL)
          {
            time1 = time2;
            if (lastSuccess == 0)
              printf("curr candidate: %ld\n", curM);
            else
              printf("last success: %ld, curr candidate: %ld\n", lastSuccess, curM);
            flushMatlab();
          }
	}
      }
    }

    cur_bitmap_size = (1+curM/(1U<<6)) * sizeof(uint64_t);
    memset(bitmap, 0, cur_bitmap_size);

    for (k = 0; k < len_freq; k++)
    {
      uint64_t index;
      index = modl(tempsp[k], curM);
        
      if (GET_BIT(bitmap, index) != 0)
        break;
      SET_BIT(bitmap, index);
    }

    if (k == len_freq)
    {
      #pragma omp critical
      {
        if (found)
	{
	  if (curM < lastSuccess)
            lastSuccess = curM;
	}
	else
          lastSuccess = curM;

	found = true;

        if (display >= DISPLAY_ITER)
        {
#ifdef _OPENMP
          printf("thread %d found lattice size: %ld\n", omp_get_thread_num(), curM);
#else
          printf("found lattice size: %ld\n", curM);
#endif
          flushMatlab();
          time1 = getcurtime();
        }
      }
    }
  }
  #pragma omp critical
  {
    mxFree(bitmap);
  }
 }

  if (interrupted)
  {
    mxFree(tempsp);
    if (prime == 1)
      mxFree(sieve);

    return -1;
  }
}

  mxFree(tempsp);
  if (prime == 1)
    mxFree(sieve);

  if (interrupted)
    return -1;

  if (display >= DISPLAY_ITER)
  {
    printf("Time: %.3e\n", getcurtime() - time0);
  }

  return lastSuccess;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  double *dfreq;
  int nfreq;
  int d;
  int64_t M;
  int64_t outM;
  int l, t;
  int *freq;
  int64_t *z;
  int display = DISPLAY_VERBOSE;
  int prime = 0;
  double *zin;
  int lzin;
  int64_t Mmin;
  int increase = 1;
#ifdef _OPENMP
  int nthreads = omp_get_max_threads();
#else
  int nthreads = 1;
#endif

  if(nrhs < 1)
    mexErrMsgIdAndTxt("gen_lattice_mex:nrhs",
                      "At least three inputs required. Usage:\nM = lattice_size_reduce_mex(frequency_index_set, lattice_size_M, z)\n  or\nM = lattice_size_reduce_mex(frequency_index_set, lattice_size_M, z, opts)\n  with possible options\n    opts.display = 'off' | 'none' | 'iter' | 'xs'\n    opts.prime = 0 | 1\n    opts.increase = 0 | 1\n    opts.Mmin = <min lattice size M>\n    opts.nthreads = <max. number of threads>");

  if(nrhs < 3)
    mexErrMsgIdAndTxt("gen_lattice_mex:nrhs",
                      "Three inputs required. Usage: M = lattice_size_reduce_mex(frequency_index_set, lattice_size_M, z)");

  if(nlhs!=1)
    mexErrMsgIdAndTxt("gen_lattice_mex:nlhs",
                      "One output required. Usage: lattice_size_reduce_mex(frequency_index_set, lattice_size_M, z)");

  if(mxGetNumberOfElements(prhs[1]) != 1 )
    mexErrMsgIdAndTxt("gen_lattice_mex:M_not_scalar",
                      "Input lattice size M (second argument) must be a scalar.");

  nfreq = mxGetM(prhs[0]);
  Mmin = 0;
  d = mxGetN(prhs[0]);
  if (mxGetScalar(prhs[1]) > (double) (1ULL<<53))
    mexErrMsgIdAndTxt("gen_lattice_mex:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^53.");
    
  M = llrint(mxGetScalar(prhs[1]));

  if (nrhs > 3 && mxIsStruct(prhs[3]))
  {
    int index = mxIsStruct(prhs[2]) ? 2 : 3;
    char str[256];
    mxArray *field;

    field = mxGetField(prhs[index], 0, "display");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Display");
    if (field != NULL && mxGetString(field, str, 255) == 0)
    {
      if (strcmp(str, "off") == 0 || strcmp(str, "Off") == 0
        ||strcmp(str, "none") == 0 || strcmp(str, "None") == 0)
        display = DISPLAY_OFF;
      else if (strcmp(str, "iter") == 0 || strcmp(str, "Iter") == 0)
        display = DISPLAY_ITER;
      else if (strcmp(str, "xs") == 0 || strcmp(str, "Xs") == 0)
        display = DISPLAY_XS;
    }

    field = mxGetField(prhs[index], 0, "prime");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Prime");
    if (field != NULL && lrint(mxGetScalar(field)) == 1)
      prime = 1;

    field = mxGetField(prhs[index], 0, "Mmin");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "mmin");
      if (field != NULL)
      {
        Mmin = llrint(mxGetScalar(field));
	if (Mmin < nfreq)
	  Mmin = nfreq;
      }

    field = mxGetField(prhs[index], 0, "increase");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Increase");
    if (field != NULL && lrint(mxGetScalar(field)) == 0)
      increase = 0;

    field = mxGetField(prhs[index], 0, "nthreads");
    if (field != NULL)
    {
#ifdef _OPENMP
      nthreads = min_i32(lrint(mxGetScalar(field)), omp_get_max_threads());
#else
      nthreads = min_i32(lrint(mxGetScalar(field)), 1);
#endif

      if (nthreads < 1)
        nthreads = 1;
    }
  }

  if (prime == 1 && M > (1ULL<<37))
    mexErrMsgIdAndTxt("gen_lattice_mex:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^37 if prime search enabled.");

  if (M < nfreq)
    mexErrMsgIdAndTxt("gen_lattice_mex:M_less_than_card",
                      "Input lattice_size_M is smaller than cardinality of frequency index set.");

  dfreq = mxGetPr(prhs[0]);

  z = (int64_t *) mxMalloc(d*sizeof(int64_t));

  lzin = mxGetN(prhs[2]);
  if (mxIsStruct(prhs[2]) || lzin != d)
    mexErrMsgIdAndTxt("gen_lattice_mex:invalid_z",
                      "Input generating vector z (third argument) must be row vector of length d");

  zin = mxGetPr(prhs[2]);
  for (t = 0; t < d; t++)
    z[t] = llrint(zin[t]);

  if (display > DISPLAY_OFF)
  {
    printf("z: ");
    for (t = 0; t < d; t++)
      printf("%ld ", z[t]);
    printf("\n");
  }

  if (Mmin < nfreq)
  {
    Mmin = nfreq;
    for (t = 0; t < d; t++)
      Mmin = max_i(Mmin, z[t]);
  }

  if (display > DISPLAY_OFF)
  {
    printf("d = %d, card_I = %d, lattice_size_M = %ld, Mmin = %ld, prime = %d, increase = %d, nthreads = %d, display = ", d, nfreq, M, Mmin, prime, increase, nthreads);
    if (display == DISPLAY_ITER)
      printf("iter");
    else if (display == DISPLAY_VERBOSE)
      printf("verbose");
    else if (display == DISPLAY_XS)
      printf("xs");
    printf("\n");
    flushMatlab();
  }


  freq = (int *) mxMalloc(d*nfreq*sizeof(int));
  for (l = 0; l < nfreq; l++)
    for (t = 0; t < d; t++)
      freq[d*l+t] = lrint(dfreq[nfreq*t+l]);

  if (increase == 1)
    outM = lattice_size_increase(d, z, freq, nfreq, M, Mmin, prime, nthreads, display);
  else
    outM = lattice_size_reduce(d, z, freq, nfreq, M, Mmin, prime, nthreads, display);

  if (outM == 0)
    outM = M;

  if (outM > 0)
  {
    double *retM;
    plhs[0] = mxCreateDoubleMatrix(1,1,mxREAL);
    retM = mxGetPr(plhs[0]);
    retM[0] = outM;

    if (display > DISPLAY_OFF)
    {
      printf("M = %ld\n", outM);
      flushMatlab();
    }
  }

  mxFree(freq);
  mxFree(z);

  if (outM == -1)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex:ctrl_c",
                      "CTRL-C pressed. Aborted.");
  }

  if (outM <= 0)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex:internal_error",
                      "No valid lattice size M found.");
  }
}

