function [ num_freq, num_freq_out, num_intersect, max_err_abs, err_l2_rel, SNR_out, SNR_out2, noise_sigma, freq_cand_sizes, lattice_sizes, sample_sizes, z, run_times ] = test4snr_sfft_algorithm_singletest( d, N, k, algorithm_name, coeff_randtype, SNR, sparsity_s, sparsity_s_local, niter_r, niter_r_basic )
%TEST4SNR_SFFT_ALGORITHM_SINGLETEST Performs a single test run of sfft_algorithm with noise.
%
%   Input parameters:
%      d - array d
%      N - search space parameter N, search space Gamma = [-N,..,N]^d
%      k - signal sparsity
%      algorithm_name - type of algorithm,
%                       either 'a1r1l' for Algorithm 1
%                       or 'a2r1l' for Algorithm 2
%      coeff_randtype - type of random coefficients,
%                       either 'rand' for fhat_k in [-1,1] + 1i*[-1,1]
%                       with |fhat_k| >= 1e-6
%                       or 'phase' for fhat_k = exp(2*pi*1i*phi_k),
%                       phi_k in [0,1)
%      SNR - (linear) signal-to-noise ratio
%      sparsity_s - (optional) target sparsity of output, default (2*N+1)^d
%      sparsity_s_local - (optional) value of sparsity_s in local loops,
%                         uses the value of sparsity_s as default
%      niter_r - (optional) number of detection iteration r in steps 2b-2e,
%                default is 1
%      niter_r_basic - (optional) number of detection iteration r in
%                      steps 1 and 2a, uses the value of niter_r as default
%
%   Output parameters:
%      num_freq - signal sparsity (input parameter k)
%      num_freq_out - number of frequencies returned by the algorithm
%      num_intersect - number of correctly detected frequencies
%      max_err_abs - maximal absolute error (incl. incorrect detection)
%      err_l2_rel - maximal relative l_2 error (incl. incorrect detection)
%      SNR_out - measured SNR value
%      SNR_out2 - measured SNR value
%      noise_sigma - standard deviation sigma of complex Gaussian noise
%      freq_cand_sizes - array of length d with number of investigated
%                        frequencies in each dimension increment step t
%      lattice_sizes - array of length d with size M of reconstructing
%                      rank-1 lattices used for the sampling
%                      in each dimension increment step t
%      sample_sizes - array of length d with number of samples taken from
%                     the function fct_handle in each dimension increment
%                     step t
%      z - generating vector z of reconstructing rank-1 lattice for
%          frequencies in freq, lattice_sizes(end) contains the
%          corresponding rank-1 lattice size M
%      run_times - array of length 5 with runtimes measured for various
%                  parts of the algorithm,
%                  run_times(1): steps 1 and 2a,
%                  run_times(2): building of reconstructing rank-1 lattices
%                  run_times(3): sampling step 2c
%                  run_times(4): inverse rank-1 lattice FFT step 2d
%                  run_times(5): total runtime of the algorithm
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

freq = unique(round((2*N+1)*rand(k,d)-(N+0.5)),'rows');
while size(freq,1) < k
  freq = union(freq, unique(round((2*N+1)*rand(k,d)-(N+0.5)),'rows'), 'rows');
end
freq = freq(1:k,:);
if strcmp(coeff_randtype, 'rand')
  fhat = 2*rand(k,1)-1 + 2i*rand(k,1)-1i;
  fhat = fhat(abs(fhat) >= 1e-6);
  while size(fhat,1) < k
    fhat = union(fhat, 2*rand(k,1)-1 + 2i*rand(k,1)-1i, 'rows');
    fhat = fhat(abs(fhat) >= 1e-6);
  end
  fhat = fhat(1:k);
elseif strcmp(coeff_randtype, 'phase')
  fhat = exp(2*pi*1i*rand(k,1));
else
  error('invalid parameter coeff_randtype with value %s', coeff_randtype);
end
threshold_theta = 1e-12;
basic_threshold_theta = 1e-12;
noise_sigma = sqrt(1/SNR)*sqrt(k);

[freq_out, fhat_out, freq_cand_sizes, lattice_sizes, sample_sizes, z, run_times] = sfft_algorithm(@(varargin) trig_poly_eval_gaussiannoise(freq, fhat, noise_sigma, varargin), d, N, algorithm_name, sparsity_s, sparsity_s_local, threshold_theta, basic_threshold_theta, niter_r, niter_r_basic );

test_signal = trig_poly_eval(freq, fhat, {z, lattice_sizes(end), []});
test_noise = complex_gaussian_noise(lattice_sizes(end),1,noise_sigma);
SNR_out = mean(abs(test_signal).^2)/mean(abs(test_noise).^2);
SNR_out2 = sum(abs(fhat).^2)/mean(abs(test_noise).^2);

num_freq = size(freq,1);
num_freq_out = size(freq_out,1);
[~, ia, ib] = intersect(freq,freq_out,'rows');
num_intersect = size(ia,1);

long_fhat = fhat(:);
long_fhat_out = zeros(size(long_fhat));
long_fhat_out(ia) = fhat_out(ib);

freq_neg = setdiff(freq_out,freq,'rows');

if ~isempty(freq_neg)
  [~, ~, ib_ianeg] = intersect(freq_neg,freq_out,'rows');
  long_fhat = [long_fhat; zeros(length(ib_ianeg),1)];
  long_fhat_out = [long_fhat_out; fhat_out(ib_ianeg)];
end

max_err_abs = max(abs(long_fhat-long_fhat_out));
err_l2_rel = norm(long_fhat-long_fhat_out,2) / norm(long_fhat,2);

end

